<?php

$store_name = "unnatural2";

$pacific_timezone = new DateTimeZone("America/Los_Angeles");
$release_date = new DateTime("2016-01-01T09:30", $pacific_timezone);
$now = new DateTime("now", $pacific_timezone);
$prerelease = ($now->getTimestamp()) < ($release_date->getTimestamp());

function json_error($msg, $code = 500) {
  if ($code == 500) {
    Header( "HTTP/1.1 500 Internal Server Error" );
  } else if ($code == 403) {
    Header( "HTTP/1.1 403 Forbidden" );
  } else if ($code == 404) {
    Header( "HTTP/1.1 404 Not Found" );
  } else if ($code == 400) {
    Header( "HTTP/1.1 400 Bad Request" );
  } else if ($code == 409) {
    Header( "HTTP/1.1 409 Conflict" );
  }
  Header( "Content-Type: application/javascript");
  
  echo json_encode(array('error' => $msg))."\n";
  throw new Exception($msg);
}

if ($store_name == "INSERTINSERTINSERT") {
  return json_error("store name not configured", 500);
}

function search_purchases($needles, $purchased) {
  $purchasedMap = array();
  foreach ($purchased as $purchase) {
    $purchasedMap[$purchase] = 1;
  }
  $results = array();
  foreach ($needles as $needle) {
    $results[$needle] = array_key_exists($needle, $purchasedMap);
  }
  return $results;
}

$full_path = $_SERVER['REQUEST_URI'];

$last_slash_index = strrpos($full_path, "/");
$queston_mark_index = strrpos($full_path, "?");
if (!$queston_mark_index) {
  $queston_mark_index = strlen($full_path);
}
$scene_filename = urldecode(substr($full_path, $last_slash_index+1, $queston_mark_index - $last_slash_index - 1));

preg_match("!(.+?)\.txt(\.json)?!", $scene_filename, $matches);
$is_download = preg_match("!\.(dmg|exe|deb)$!", $scene_filename);
$is_omnibus_app = preg_match("!CoGnibus!", $_SERVER['HTTP_USER_AGENT']);
$is_android = preg_match("!Android!", $_SERVER['HTTP_USER_AGENT']);
$scene_name = $matches[1];

$debug_log = 0;

if ($debug_log) {
  error_log("full_path $full_path");
  error_log("store_name $store_name");
  error_log("scene_filename $scene_filename");
  error_log("scene_name: $scene_name");
  error_log("is_download: $is_download");
  error_log("is_omnibus_app: $is_omnibus_app");
  error_log("is_android: $is_android");
}

if (!file_exists("mygame.js")) {
  return json_error("couldn't find mygame", 404);
}

if (!file_exists("scenes/$scene_filename")) {
  return json_error("couldn't find scene", 404);
}

$contents = file_get_contents("scenes/$scene_filename");
$file_hash = md5($contents);

if (array_key_exists("hash", $_GET)) {
  $hash = $_GET['hash'];
  if ($hash != $file_hash) {
    return json_error("hash doesn't match", 404);
  }
} else {
  $hash = 0;
}

$mygame = file_get_contents("mygame.js");
$beta = str_contains($mygame, "\nbeta = \"beta\";");
if ($beta) {
  $purchases = [];
} else {
  preg_match("/purchases =\s*({[^}]*})/s", $mygame, $matches);
  $purchases = json_decode($matches[1], true);
}

$ok = false;

$requires_purchase = $purchases && array_key_exists($scene_name, $purchases);
$requires_purchase = $requires_purchase || $is_download;

$skip_purchase = $is_omnibus_app && !$is_android;

if ($debug_log) {
  error_log("purchases: " . json_encode($purchases));
  error_log("requires_purchase: $requires_purchase");
  error_log("skip_purchase: $skip_purchase");
}

if ($requires_purchase && !$skip_purchase) {
  session_name("cog");
  session_set_cookie_params(0,'/',"",true,true);
  session_start();

  $product = $store_name.".".$purchases[$scene_name];
  if (array_key_exists("steam_purchases", $_SESSION)) {
    foreach ($_SESSION["steam_purchases"] as $purchase) {
      if ($product == $purchase) {
        $ok = true;
        break;
      }
    }
  }
  if (!$ok && array_key_exists("purchases", $_SESSION)) {
    $purchased = $_SESSION["purchases"];

    foreach ($purchased as $purchase) {
      if ($product == $purchase) {
        $ok = true;
        break;
      }
    }
  }
  if ($debug_log) {
    error_log("steam_purchases: " . $_SESSION["steam_purchases"]);
    error_log("session purchases: " . json_encode($_SESSION["purchases"]));
    error_log("product: $product");
    error_log("ok: $ok");
  }
  if (!$ok) {
    if (!array_key_exists("login_id", $_SESSION)) json_error("not registered", 403);

    if ($prerelease && $_SESSION['email'] != 'fullaccess@choiceofgames.com') {
      json_error("prerelease", 403);
    }

    error_log("not purchased 403 " . $product . ' ' . $_SESSION['login_id'] . ' ' . $_SESSION['email']);
    json_error("not purchased", 403);
  }
}

if (preg_match("/\.txt$/", $scene_filename)) {
  header("Content-Type: text/plain");
} else if (preg_match("/\.json$/", $scene_filename)) {
  header("Content-Type: application/json");
} else if (preg_match("/\.(exe|deb|dmg)$/", $scene_filename)) {
  header("Content-Type: application/octet-stream");
}

if ($hash && !$requires_purchase) {
  header("Cache-Control: max-age=31536000");
} else {
  header("Cache-Control: max-age=0, must-revalidate");
}

$etag = "W/\"$file_hash\"";
header("ETag: $etag");

if (isset($_SERVER['HTTP_IF_NONE_MATCH']) && $etag == $_SERVER['HTTP_IF_NONE_MATCH']) {
  header( "HTTP/1.1 304 Not Modified" );
  return;
}

echo $contents;

?>